package joystick;

import java.util.ArrayList;
import java.util.logging.Level;
import java.util.logging.Logger;

import test.TopicTest;

import dcps.Topic;
import dcps.TopicsScope;
import ethernet.MultiPeer;

import net.java.games.input.Component;
import net.java.games.input.Controller;
import net.java.games.input.ControllerEnvironment;
import net.java.games.input.Component.Identifier;

public class JoystickDCPS {
	
	private ArrayList<Controller> foundControllers;
	private Topic topic = null;
	
    public JoystickDCPS(Topic to) {       
    	topic = to;
        foundControllers = new ArrayList<>();
        searchForControllers();
        // If at least one controller was found we start the sampling thread.
        if(!foundControllers.isEmpty()) {
            new SamplingThread().start();
        } else {
            System.out.println("No controller found!");
        }
    }

    /**
     * Search (and save) for controllers of type Controller.Type.STICK,
     * Controller.Type.GAMEPAD, Controller.Type.WHEEL and Controller.Type.FINGERSTICK.
     */
    private void searchForControllers() {
        Controller[] controllers = ControllerEnvironment.getDefaultEnvironment().getControllers();
        for(int i = 0; i < controllers.length; i++){
            Controller controller = controllers[i];
            if (
                    controller.getType() == Controller.Type.STICK || 
                    controller.getType() == Controller.Type.GAMEPAD || 
                    controller.getType() == Controller.Type.WHEEL ||
                    controller.getType() == Controller.Type.FINGERSTICK
               )
            {
            // Add new controller to the list of all controllers.
            foundControllers.add(controller);       
            // Add new controller to the list on the window.
            System.out.println(controller.getName() + " - " + controller.getType().toString() + " type");
            }
        }
    }
    
	
	private class SamplingThread extends Thread {
		@Override
		public void run() {	
			String msg = new String();
			String prevMsg = new String();
			int count = 0;
			while(Thread.currentThread().isAlive()) {
		        	//msg = "Message: ";
		        	JoystickData joyData = new JoystickData();
		            // Currently selected controller.
		            int selectedControllerIndex = 0; 
		            Controller controller = foundControllers.get(selectedControllerIndex);
		            // Pull controller for current data, and break while loop if controller is disconnected.
		            if( !controller.poll() ){
		                break;
		            }	            
		            // X axis and Y axis
		            int xAxisPercentage = 0;
		            int yAxisPercentage = 0;		                    
		            // Go trough all components of the controller.
		            Component[] components = controller.getComponents();
		            for(int i=0; i < components.length; i++)
		            {
		                Component component = components[i];
		                Identifier componentIdentifier = component.getIdentifier();
		                
		                //System.out.println("Joy component: "+ componentIdentifier.getName());
		                // Buttons
		                if(componentIdentifier.getName().matches("^[0-9]*$")){ // If the component identifier name contains only numbers, then this is a button.
		                    // Is button pressed?
		                    boolean isItPressed = true;
		                    if(component.getPollData() == 0.0f)
		                        isItPressed = false;         
		                    // Button index
		                    //String buttonIndex;
		                    //buttonIndex = component.getIdentifier().toString();
		                    // Add to the message
		                   // msg += " "+buttonIndex + " " + isItPressed;
		                    if (i==0) joyData.setButton0(isItPressed);
		                    if (i==1) joyData.setButton1(isItPressed);
		                    if (i==2) joyData.setButton2(isItPressed);
		                    if (i==3) joyData.setButton3(isItPressed);
		                    // We know that this component was button so we can skip to next component.
		                    continue;
		                }
		                
		                // Hat switch
		                if(componentIdentifier == Component.Identifier.Axis.POV){
		                    float hatSwitchPosition = component.getPollData();
		                    // Add to the message
		                    //msg += " "+hatSwitchPosition;
		                    joyData.setHat(hatSwitchPosition);
		                    // We know that this component was hat switch so we can skip to next component.
		                    continue;
		                }
		                
		                // Axes
		                if(component.isAnalog()){
		                    float axisValue = component.getPollData();
		                    int axisValueInPercentage = getAxisValue(axisValue);
		                    
		                    // X axis
		                    if(componentIdentifier == Component.Identifier.Axis.X){
		                        xAxisPercentage = axisValueInPercentage;
		                        continue; // Go to next component.
		                    }
		                    // Y axis
		                    if(componentIdentifier == Component.Identifier.Axis.Y){
		                        yAxisPercentage = axisValueInPercentage;
		                        continue; // Go to next component.
		                    }
		                    // Add to the message
		                    //msg += " "+axisValueInPercentage;
		                }
		            }

		            // Add to the message
		            //msg += " Axis: "+xAxisPercentage + " " + yAxisPercentage;
		            joyData.setStickPosX(xAxisPercentage);
		            joyData.setStickPosY(yAxisPercentage);

		            msg = joyData.toString();
		            
		            if (!msg.contentEquals(prevMsg)) {
			            count = (count+1)%80;
			            if (count != 0) 
			            	System.out.print(".");
			            else System.out.println(".");
			          //publish
		            	topic.publish(msg.getBytes());
		            }
		            prevMsg = msg;
		            //System.out.println(msg);          
		            // We have to give processor some rest.
		            try {
		                Thread.sleep(50);
		            } catch (InterruptedException ex) {
		            	ex.printStackTrace();
		            }

			}
		}
	}
	
    /**
     * Given value of axis in percentage.
     * Percentages increases from left/top to right/bottom.
     * If idle (in center) returns 500, if joystick axis is pushed to the left/top 
     * edge returns 0 and if it's pushed to the right/bottom returns 1000.
     * 
     * @return value of axis in percentage.
     */
    public static int getAxisValue(float axisValue)
    {
        return (int)(((2 - (1 - axisValue)) * 1000) / 2);
    }
    
    public static void main(String[] args) {
    	
		MultiPeer multiPeer = new MultiPeer();
		TopicsScope ts = new TopicsScope(multiPeer);
		Topic t1 = ts.lookup("T01");
		new JoystickDCPS(t1);
		
    }

}
