
package ethernet;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.net.SocketAddress;
import java.net.SocketTimeoutException;
import java.net.UnknownHostException;
/**
 * 
 * The class Destination represents a functional TCP/IP bidirectional channel.
 * A member IPPort object represents the end point identity.
 * When created, a Destination is not functional, it must be consolidated by 
 * an "accept" or "connect" operation. The parent "MultiPeer" (that holds a 
 * list of Destinations) is the responsible of these operations.
 * Each instance of Destination has an independent receiving thread that
 * notify to the parent MultiPeer a message arrival (calling back its 
 * "OnMesssage" method).
 * 
 * @author Jose Simo (c) ai2-UPV 2014
 * 
 */
public class Destination {
	
	// milliseconds to set the timeout of socket connect
	private static int TIMEOUT_CONNECTION = 300;  
	
	private IPPort ipport = null;
	private Socket socket = null;
	private OutputStream os = null;
	private InputStream is = null;
	private boolean connected = false;
	private MultiPeer multiPeerParent = null;
	
	public IPPort getIPPort() {return ipport;}
	
	/**
	 * Create a Destination instance linked to a MultiPeer object.
	 * @param multipeer
	 */
	Destination(MultiPeer multipeer) {
		this.multiPeerParent = multipeer;
	}

	/**
	 * Connects to the destination end point represented by the passed IPPort object.
	 * Create the input/output streams artifacts.
	 * In order to inform to the end point peer about the caller identity, it sends 
	 * a string representation of the local peer identity (IPPort) at the very 
	 * beginning of the communication.
	 * Once the connection is established, it start its private receiving thread.
	 * @param ipp
	 * @return true if connected, false if the Destination channel is not functional.
	 */
	public synchronized boolean connect(IPPort ipp) {
		boolean retval = true;
		//If already connected the call does not have any effect.
		if (connected) return retval;
		//Set the IPPort end point identity
		this.ipport = ipp;
		try {
			//Connect to the end point
			SocketAddress sockAddr = new InetSocketAddress(ipport.getAddress(),ipport.getPort());
			socket = new Socket();
			socket.connect(sockAddr, TIMEOUT_CONNECTION);
			//Disable the Nagle's algorithm (optimization of the number of packets in the wire)
			//this reduce the latency of the link.
			socket.setTcpNoDelay(true); 
			//Create in/out streams
			os = socket.getOutputStream();
			is = socket.getInputStream();
			//// Send to the "accepting" peer the local identity
				String m = multiPeerParent.getLocalIPPort().toString();
				byte[] buf = m.getBytes();
				byte[] sizeBuff = util.Bin.intToDWord(buf.length);
				//Write the size (4bytes) and the IPPort (String)
				os.write(sizeBuff, 0, 4);
				os.write(buf, 0, buf.length);
			//// Well, the connection is ready
			connected = true;
			// Starts the reading thread
			new DestinationReadThread().start();
			util.Log.logMessage(util.Log.LogType.INFO,"Destination: Connected (connect) to: "+ ipport.toString());
		} catch (SocketTimeoutException e) {
			close();
			retval = false;
			util.Log.logMessage(util.Log.LogType.ERROR,"Destination: Connect: Unreachable endpoint. Timeout : "+ipport.toString());
			//e.printStackTrace();
		} catch (UnknownHostException e) {
			close();
			retval = false;
			util.Log.logMessage(util.Log.LogType.ERROR,"Destination: Connect: Unknown host: "+ipport.toString());
			//e.printStackTrace();
		} catch (IOException e) {
			util.Log.logMessage(util.Log.LogType.ERROR,"Destination: Connect: Connection error: "+ipport.toString());	
			retval = false;
			close();
			//e.printStackTrace();
		}
		return retval;
	}
	
	/**
	 * Consolidate the Destination instance around an "accepted" connection.
	 * Create the input/output streams artifacts.
	 * At the very begging of the communications, it reads the peer identity from the input stream.
	 * Once the connection is established, it start its private receiving thread.
	 * @param sock is the socket obtained from an "accept" operation (performed in the MultiPeer parent).
	 * @return true if connected, false if the Destination channel is not functional.
	 */
	public synchronized boolean accepted(Socket sock) {
		boolean retval = true;
		//If already connected the call does not have any effect.
		if (connected) return retval;
		socket = sock;
		try {
			//Disable the Nagle's algorithm (optimization of the number of packets in the wire)
			//this reduce the latency of the link.
			socket.setTcpNoDelay(true); 
			//Create in/out streams
			os = socket.getOutputStream();
			is = socket.getInputStream();
		    ////Read from the input stream the peer identity.
			//Read the identity size
			byte[] sizeBuf = new byte[4];
			is.read(sizeBuf, 0, 4);
			int size = util.Bin.dWordToInt(sizeBuf);
			//Read the identity
			byte[] data = new byte[size];
			is.read(data, 0, size);
			String m = new String(data);
			//Parse the received IPPort
			IPPort ippReceived = IPPort.parseIPPort(m);
			if (ippReceived != null) {
			    //Set the IPPort end point identity
				ipport = ippReceived;
			    // Well, the connection is ready
				connected = true;
				// Starts the reading thread
				new DestinationReadThread().start();
				util.Log.logMessage(util.Log.LogType.INFO,"Destination: Connected (accepted) from: "+ ipport.toString());
			}
		} catch (IOException e) {
			util.Log.logMessage(util.Log.LogType.ERROR,"Destination: Accepted: Connection error: "+ipport.toString());
			retval = false;
			close();
			//e.printStackTrace();
		}
		return retval;
	}	
	
	/**
	 * Send a message to the peer (header and payload buffers of the Message passed)
	 * The information about message length and type is encoded in a prologue handled 
	 * at this level (Destination.send() and Destination.receive()). The prologue is:
	 * [Message Length (4 bytes)] + [Message Type (1 byte)]
	 * @param m the Message to send.
	 * @return true if the message has been sent successfully, false if communication error.
	 */
	public synchronized boolean send(Message m) {
		//If not connected the call does not have any effect.
		if (!connected) return false;
		//If there is not a message to send, returns
		if (m == null) return false;
		//Let's write!
		try {
			m.toOutputStream(os);
			//force output stream flush
			os.flush();
		} catch (IOException e) {
			util.Log.logMessage(util.Log.LogType.ERROR,"Destination: Write error: "+ipport.toString());
			if (multiPeerParent != null) {
				//If the write fails, remove this destination from the parent MultiPeer list.
				multiPeerParent.removeDestination(this);
			}
			close();
			return false;
		}
		return true;
	}
	
	/**
	 * Private reading routine. 
	 * It is called from the "DestinationReadThread" to propagate the received messages 
	 * to the parent MultiPeer.
	 * This method reads the prologue from the input stream and creates an object matching 
	 * the received message type. 
	 * @return the received message; null, if communication error.
	 */
	private Message receive() {
		//If not connected the call does not have any effect.
		if (!connected) return null;
		Message m = null;		
		try {
			m = Message.fromInputStream(is);
		} catch (IOException e) {
			util.Log.logMessage(util.Log.LogType.ERROR,"Destination: Read error: "+ipport.toString());
			close();
			return null;
		}
		return m;
	}
	
	/**
	 * Private "close" operation.
	 */
	private void close() {
		try {
			if (socket != null) socket.close();
			if (os != null) os.close();
			if (is != null) is.close();
			socket = null; os = null; is = null; //ipport = null;
			connected = false;
		} catch (IOException e) {
			util.Log.logMessage(util.Log.LogType.ERROR,"Destination: Closing error: "+ipport.getAddress());
			e.printStackTrace();
		}
	}

	/**
	 * This private thread runs forever reading messages by calling "receive" method.
	 * Once message received, it is passed to the MultiPeer parent by calling-back
	 * MultiPerr.OnMessage 
	 * @author jsimo
	 *
	 */
	private class DestinationReadThread extends Thread {
		@Override
		public void run() {
			 while((Thread.currentThread().isAlive()) && connected) {
				 //Receive message
				 Message m = receive();
				 if (m == null) {
					 if (multiPeerParent != null) {
						//If the receive fails, remove this destination from the parent MultiPeer list
						multiPeerParent.removeDestination(Destination.this);
					 }
					 break;
				 }
				 if (multiPeerParent != null) {
					 //Pass the message to the parent MultiPeer
					 multiPeerParent.OnMessage(m);
				 }
			 }
		}
	}
}
