
package ethernet;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.InetAddress;
import java.net.MulticastSocket;
import java.net.UnknownHostException;
import java.util.ArrayList;

/**
 * A Discovering Service feeds with new peers the associated MultiPeer instance.<p>
 * 
 * It uses a diffusion channel to publish the (local) peer identity 
 * (IP and port as String xx.xx.xx.xx:pppp). This publication is done in the "DiscoverWriteThread".
 * The "DiscoverReadThread" reads the diffusion channel in order to know all peers in the party.
 * New peers are notified to the associated MultiPeer by calling "addDestination". <p>
 * 
 * The diffusion channel must be agreed by all peers in the party. In fact, it identifies the group.
 * The associated MultiPeer and Diffusion Channel are provided to the DiscoveringService constructor.<p>
 * 
 * The period of the "DiscoverWriteThread" is variable belonging to 
 * [periodDiffusionMAX..periodDiffusionMIN]. It is more frequent at startup time and 
 * it is reactive when a new peer is detected.<p>
 * 
 * Both discovering threads are high-priority in order to configure the system as soon as possible.<p>
 * 
 * @author Jose Simo (c) ai2-UPV 2014
 */
public class DiscoveringService {

	private static int MAX_NET_TRIES = 5;
	private static int periodDiffusionMAX = 10000;  // 10 seconds
	private static int periodDiffusionMIN = 100;    // 0.1 seconds
	private int periodDiffusion = periodDiffusionMIN;
	
	
	private IPPort muticastIPPort = null;
	private MulticastSocket socket = null;	
	private InetAddress group = null;
	private MultiPeer multiPeer = null;
	private int newPeerDetectedCounter = 0;
	private boolean connected = false;	
	/**
	 * @return true if service connected.
	 */
	public boolean isConnected() {return connected;}
	

	/**
	 * DiscoveringService constructor. It starts the service.
	 * @param mp is the associated MultiPeer. It should not be null.
	 * @param ipp is the diffusion channel. Multicast IP and any port.
	 */
	DiscoveringService(MultiPeer mp, IPPort ipp) {		
		multiPeer = mp;
		muticastIPPort = ipp;
		startDiscoveringService();
	}
	
	/**
	 * Create the communications artifacts and starts the working threads
	 */
	private void startDiscoveringService() {
		//Create the diffusion group
		try {
			group = InetAddress.getByName(muticastIPPort.getAddress().getHostAddress());
		} catch (UnknownHostException e2) {
			e2.printStackTrace();
			return;
		}
		//Create the multicast socket
		try {
			socket = new MulticastSocket(muticastIPPort.getPort());
			socket.joinGroup(group);
		} catch (IOException e1) {
			e1.printStackTrace();
			return;
		}
		connected = true;
		//Create and starts the working threads.
		
		Thread readTh = new DiscoverReadThread();
		readTh.setPriority(Thread.MAX_PRIORITY);
		readTh.start();
		Thread writeTh = new DiscoverWriteThread();
		writeTh.setPriority(Thread.MAX_PRIORITY);
		writeTh.start();
		
		/*
		Thread staticTh = new StaticPeersThread();
		staticTh.setPriority(Thread.MAX_PRIORITY);
		staticTh.start();*/
	}
	
	/**
	 * Stop threads and close communication artifacts.
	 */
	private void stopDiscoveringService() {
		connected = false;
		if (socket != null) {
			socket.close();
		}
	}
	/**
	 * Used by the "DiscoverWriteThread" to control the execution period.
	 * It waits to the detection of a new peer using a variable timeout.
	 * This mechanism ensure a "shot time" identity announcement at startup, 
	 * a "long time" announcement when the system is stable and a reactive 
	 * one when a new peer is detected.
	 */
	private synchronized void waitToNextLoop() {
		if (newPeerDetectedCounter <= 0) {
			try {
				this.wait(periodDiffusion);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
		newPeerDetectedCounter--;
		periodDiffusion = (periodDiffusion >= periodDiffusionMAX) ? periodDiffusionMAX: periodDiffusion + periodDiffusion/2;
	}
	/**
	 * Used by "DiscoverReadThread" in order to notify the detection of a new peer.
	 */
	private synchronized void notifyIPPortReception(IPPort ipPort) {
			periodDiffusion = periodDiffusionMIN;
			newPeerDetectedCounter++;
			this.notify();
	}
	
	/**
	 * @author jsimo
	 * This thread reads the diffusion channel in order to discover new peers.
	 */
	private class DiscoverReadThread extends Thread {
		@Override
		public void run() {		
			byte[] buffer = new byte[256];;
			DatagramPacket packet = new DatagramPacket(buffer, buffer.length);;
		    int tries = 0;
		    while(Thread.currentThread().isAlive() && connected) {
			    try {
			    	//Read from the channel
			    	//Clean the buffer
			    	for(int i=0;i<buffer.length;i++) buffer[i] = 0;
			    	//Receive
					socket.receive(packet);
					tries = 0;
				} catch (IOException e1) {
					util.Log.logMessage(util.Log.LogType.WARNING,"Discovering Service: Read Thread: Network error. Retrying...");
					tries++;
					//If too many errors, stop the service and exit the thread
					if (tries > MAX_NET_TRIES) {
						util.Log.logMessage(util.Log.LogType.ERROR,"Discovering Service: Severe network error. Exit reading thread.");
						stopDiscoveringService();
						return;
					}
					continue;
				}
			    //Extract the buffer read.
			    String mensaje = new String(packet.getData());
			    IPPort receivedPeerIPPort = IPPort.parseIPPort(mensaje);	
	 //System.out.println("Discover received: "+ mensaje + " -- "+ receivedPeerIPPort.getPort());
			    //The associated MultiPeer should not be null
			    if (multiPeer != null) {
				    boolean isMyEcho = receivedPeerIPPort.isEqual(multiPeer.getLocalIPPort());
				    if (isMyEcho) {    
				    	//The read is my own identity. Do nothing.
				    } else {
						//Only new destinations are taken into account.
						if (!multiPeer.isDestinationContained(receivedPeerIPPort)) {
							//Notify a new peer found
					    	notifyIPPortReception(receivedPeerIPPort);   //To the "DiscoverWriteThread"
					    	multiPeer.addDestination(receivedPeerIPPort);//To the associated MultiPeer
						}
				    }
			    }
			}
		    util.Log.logMessage(util.Log.LogType.WARNING,"Discovering Service: Exit reading thread.");
		}	
	}
	
	/**
	 * @author jsimo
	 * This thread publish the (local) peer identity (IP and port as String xx.xx.xx.xx:pppp)
	 * through the diffusion channel.
	 */
	private class DiscoverWriteThread extends Thread {
		@Override
		public void run() {			
			byte[] buffer = null; //new byte[512];
		     
		    int tries = 0;
		    while(Thread.currentThread().isAlive() && connected) {
		    	//The associated MultiPeer should not be null
		    	if (multiPeer != null) {
		    		//build the message and pack it.
					String mensaje = multiPeer.getLocalIPPort().toString();
	// System.out.println("Discover sending: "+ mensaje );
					buffer = mensaje.getBytes();
					DatagramPacket packet = new DatagramPacket(buffer, buffer.length, group, muticastIPPort.getPort()); 
				    try {
				    	//Send the packet
						socket.send(packet);
						tries = 0;
					} catch (IOException e) {
						util.Log.logMessage(util.Log.LogType.WARNING,"Discovering Service: Write Thread: Network error. Retrying...");
						tries++;
						//If too many errors, stop the service and exit the thread
						if (tries > MAX_NET_TRIES) {
							util.Log.logMessage(util.Log.LogType.ERROR,"Discovering Service: Severe network error. Exit writing thread.");
							stopDiscoveringService();
							return;
						}
					}
		    	}
		    	//Wait until the next announcement.
		    	waitToNextLoop();
			}
		    util.Log.logMessage(util.Log.LogType.WARNING,"Discovering Service: Exit writing thread.");
		}
	}

	private class StaticPeersThread extends Thread {
		@Override
		public void run() {			
			
			ArrayList<String> peersList = new ArrayList<String>();
			String strIP = multiPeer.getLocalIPPort().getAddress().getHostAddress();
			peersList.add(strIP+":5000");
			peersList.add(strIP+":5001");
			peersList.add(strIP+":5002");
			peersList.add(strIP+":5003");
			peersList.add(strIP+":5004");
		    while(Thread.currentThread().isAlive() && connected) {
		    	//The associated MultiPeer should not be null
		    	if (multiPeer != null) {
		    		for (String peer : peersList) {
			    		IPPort staticIPPort = IPPort.parseIPPort(peer);
			    		boolean isMyIdentity = staticIPPort.isEqual(multiPeer.getLocalIPPort());
			    		if (!isMyIdentity) {
			    			multiPeer.addDestination(staticIPPort);
			    		}
		    		}
		    	}
		    	
		    	try {
					Thread.sleep(2000);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}

			}
		    util.Log.logMessage(util.Log.LogType.WARNING,"Discovering Service: Exit static peers thread.");
		}
	}
}
