
package ethernet;

import java.net.InetAddress;
import java.net.UnknownHostException;
/**
 * This class encapsulates a peer identity (IP and port).
 * It internally computes a GUID with ordering properties.
 * Some utility methods are also provided.
 * 
 * @author Jose Simo (c) ai2-UPV 2014
 * 
 */
public class IPPort {
	private String host = "";
	private int port = 0;
	private InetAddress address = null;
	private long GUID = 0;
	private long ipAddressAsLong = 0;
	
	public int getPort() { return port; }
	public InetAddress getAddress() { return address; }
	public long getGUID() { return GUID; }
	public long getIpAddressAsLong() { return ipAddressAsLong; }
	
	/**
	 * Constructor
	 * @param ip IP address as String
	 * @param p port
	 */
	IPPort(String ip, int p) {
		host = ip;
		port = p;
		try {
			address = InetAddress.getByName(host);
		} catch (UnknownHostException e) {
			util.Log.logMessage(util.Log.LogType.ERROR,"IPPort: Unknown host: "+host);
			//e.printStackTrace();
		}
		ipAddressAsLong = util.Net.ipAddrAsLong(address);
		GUID = computeGUID(this);
	}
	
	/**
	 * Constructor
	 * @param ip IP address as InetAddress
	 * @param p port
	 */
	IPPort(InetAddress ip, int p) {
		host = ip.toString();
		port = p;
		address = ip;
		ipAddressAsLong = util.Net.ipAddrAsLong(address);
		GUID = computeGUID(this);
	}
	
	/**
	 * 
	 * @param ip
	 * @param p
	 */
	IPPort(byte[] ip, int p) {
		//TODO: check the buffer size (must be 4)
		try {
			address = InetAddress.getByAddress(ip);
		} catch (UnknownHostException e) {
			util.Log.logMessage(util.Log.LogType.ERROR,"IPPort: Unknown host: " + ip.toString());
			e.printStackTrace();
		}
		host = address.getHostAddress();
		port = p;
		ipAddressAsLong = util.Net.ipAddrAsLong(address);
		GUID = computeGUID(this);
	}
	
	/**
	 * Constructor (copy)
	 * @param ipp Other peer.
	 */
	IPPort(IPPort ipp) {
		this.host = ipp.host;
		this.port = ipp.port;
		this.address = ipp.address;
		this.ipAddressAsLong = ipp.ipAddressAsLong;
		this.GUID = ipp.GUID;
	}
	
	/**
	 * Gets the String form.
	 */
	public String toString() {
		return address.getHostAddress() + ":" + port; 
	}

	/**
	 * Comparison methods
	 */
	public boolean isEqual(IPPort p) {
		return GUID == p.getGUID();
	}
	
	public boolean isGreater(IPPort p) {
		return GUID > computeGUID(p);
	}
	
	public boolean isLower(IPPort p) {				
		return GUID < computeGUID(p);
	}
	
	///////////Static members	
	/**
	 * Gets a new IPPort object from an String representation.
	 * @param s
	 * @return a new IPPort object from an String representation.
	 */
	public static IPPort parseIPPort(String s) {
	    String ip = s.substring(0,s.indexOf(":")).trim();
	    String port = s.substring(s.indexOf(":")+1).trim();
	    InetAddress addr = null;
	    int portInt = 0;
	    try {
	       portInt = Integer.parseInt(port,10);
	    } catch (NumberFormatException e) {
	    	util.Log.logMessage(util.Log.LogType.ERROR,"ParseIPPort: Format error: "+ s);
			return null;
	    }
	    try {
			addr = InetAddress.getByName(ip);
		} catch (UnknownHostException e) {
			util.Log.logMessage(util.Log.LogType.ERROR,"ParseIPPort: Unknown Host: "+ s);
			return null;
		}
	    return new IPPort(addr,portInt);
	}
	
	/**
	 * Compute the GUID of a IPPort.
	 * @param p
	 * @return the GUID long value
	 */
	private static long computeGUID(IPPort p) {
		byte[] addrBuff = p.getAddress().getAddress();
		byte[] guidBuff = new byte[8];
		guidBuff[0] = addrBuff[3];
		guidBuff[1] = addrBuff[2];
		guidBuff[2] = addrBuff[1];
		guidBuff[3] = addrBuff[0];
		guidBuff[4] = 0; guidBuff[5] = 0; guidBuff[6] = 0; guidBuff[7] = 0;
		long v1 = util.Bin.qWordToLong(guidBuff);
		long retval = ( v1 << 16) + p.getPort();
		return retval;
	}
	
	////////////////
	
	/*
	public static void main(String[] args) {
		 
		 IPPort p1 = new IPPort("1.10.1.1",3);
		 IPPort p2 = new IPPort("1.10.1.1",2);
		 
		 System.out.println("Greater: "+ p1.isGreater(p2));
		 System.out.println("Lower: "+ p1.isLower(p2));
		 System.out.println("Equal: "+ p1.isEqual(p2));
		 
	}
	*/
}
