
package ethernet;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.InetAddress;

/**
 * Abstract class implementing "Message" common properties.
 * The message content depends on the type of the message, but it is
 * split in three buffers: the prologue, the header and the pay load. This separation
 * simplifies the message management.<p>
 * Each MessageType must have a specific derivation of the Message abstract class.<p>
 * 
 * The "prologue" information is managed by this parent class "Message" in the 
 * "fromInputStream" static method, and "computePrologue" instance method.<p>
 * The prologue format is:<p>
 * [0xF1 (Flag 2 bytes)] + [Message Length (4 bytes)] + [Message Type (1 byte)] + [Message Origin IP (4 bytes)] + [Message Origin port (2 bytes)]<p>
 * 
 * Messages must be received using the static method "fromInputStream". It calls inside 
 * its code the "fromInputStream" of a specific instance depending of the message type.<p>
 * 
 * @author Jose Simo (c) ai2-UPV 2014
 * 
 */

public abstract class Message {
	//Message types
	public enum MessageType {PLAIN,TOPICDATA,TOPICLINKS,SYNCTIME,PROXYWRAPPER};
	//Prologue flag.
	protected static byte[] msgFlag = {0xF,0x1};
	//Message contents
	protected byte[] bufferPrologue = null;
	protected byte[] bufferHeader = null;
	protected byte[] bufferPayload = null;
	//The message timeStamp is the local clock time of message instance creation.
	//It must be managed inside any constructor of "Message" extension.
	protected long timeStamp = 0;
	//Message type
	protected MessageType type = MessageType.PLAIN;
	//Source of the message. 
	protected IPPort sourceIPPort = null;
	public IPPort getSourceIPPort() {return sourceIPPort;}

	
	public byte[] getBufferPayload() {return bufferPayload;}
	public byte[] getBufferHeader() {return bufferHeader;}
	public byte[] getBufferPrologue() {return bufferPrologue;}
	public long getTimeStamp() {return timeStamp;}
	public MessageType getType() {return type;}
	
	/**
	 * This method populates the "MessageXXX" instance reading data from an
	 * input stream. The number of bytes to read must be provided.
	 * This method must be written for each extension of "Message" class.
	 * @param is
	 * @param size
	 * @return true if success
	 * @throws IOException
	 */
	protected abstract boolean fromInputStream(InputStream is, int size) throws IOException;
	
	/**
	 * It writes the "MessageXXX" instance to the specified stream.
	 * This method must be written for each extension of "Message" class.
	 * @param os
	 * @throws IOException
	 */
	protected abstract void toOutputStream(OutputStream os) throws IOException;
	
	/**
	 * Static method to create a new specific "MessageXXX" according the message 
	 * type contained in the prologue.
	 * Use this method to read messages from a communications channel.
	 * @param is
	 * @return a new Message instance
	 * @throws IOException
	 */
	public static Message fromInputStream(InputStream is) throws IOException {
		int nRead = 0;
		//Allocate the prologue information buffers
		byte[] flagBuff = new byte[2];
		byte[] totalSizeBuff = new byte[4];
		byte[] typeCodeBuff = new byte[1];
		byte[] sourceIPBuff = new byte[4];
		byte[] sourcePortBuff = new byte[2];
		Message m = null;	
		//Read the message flag and check it.
		nRead = is.read(flagBuff, 0, 2);	    
	    if (nRead == -1) throw new IOException();
	    if ((flagBuff[0] != msgFlag[0]) || (flagBuff[1] != msgFlag[1])) {
	    	//System.out.println("Wrong flag");
	    	throw new IOException();
	    }
		//Read the "message size" buffer
	    nRead = is.read(totalSizeBuff, 0, 4);	    
	    if (nRead == -1) throw new IOException();
	    //Read the "message type" buffer
	    nRead = is.read(typeCodeBuff, 0, 1);
	    if (nRead == -1) throw new IOException();
	    //Read the "message source IP" buffer
	    nRead = is.read(sourceIPBuff, 0, 4);
	    if (nRead == -1) throw new IOException();
	    //Read the "message source Port" buffer
	    nRead = is.read(sourcePortBuff, 0, 2);
	    if (nRead == -1) throw new IOException();  
	    //Parse the prologue buffers
		int size = util.Bin.dWordToInt(totalSizeBuff);
		Message.MessageType type = null;
		try {
			type = Message.MessageType.values()[typeCodeBuff[0]];
		} catch(ArrayIndexOutOfBoundsException e) {
			throw new IOException();
		}
		InetAddress sourceIP = InetAddress.getByAddress(sourceIPBuff);
		int sourcePort = util.Bin.wordToInt(sourcePortBuff);
		//Create a MessageXXX object according the received type
		switch (type) {
			case PLAIN:
				m = new MessagePlain();
			break;
			case TOPICDATA:
				m = new MessageTopicData();
			break;
			case TOPICLINKS:
				m = new MessageTopicLinks();
			break;
			case SYNCTIME:
				m = new MessageSyncTime();
			break;
			case PROXYWRAPPER:
				m = new MessageProxyWrapper();
			break;
		}
		//Store the source
		m.sourceIPPort = new IPPort(sourceIP, sourcePort);
		//Store the prologue buffer
		m.bufferPrologue = util.Bin.concatBuffers(flagBuff,totalSizeBuff,typeCodeBuff,sourceIPBuff,sourcePortBuff);
		//Populate the message by passing the stream to the specific instance.
		boolean success = m.fromInputStream(is, size);
		if (!success) throw new IOException();
		//System.out.println("Message received: type: " + m.getType().toString());
		return m;
	}
	
	/**
	 * Helper method to compute the header of an already populated message.
	 * 
	 * @return the prologue buffer.
	 */
	public byte[] computePrologue() {
		//Get the message header and payload buffers
		byte[] payLoad = getBufferPayload();
		byte[] header = getBufferHeader();
		//Compute the buffers length
		int payLoadSize = (payLoad!=null) ? payLoad.length : 0;
		int headerSize = (header!=null) ? header.length : 0;
		int totalSize = headerSize + payLoadSize;
		//Prepare the "message length" buffer
		byte[] totalSizeBuff = util.Bin.intToDWord(totalSize);
		//Prepare the "message type" buffer
		Message.MessageType type = getType();
		byte[] typeCodeBuff = new byte[1];
		typeCodeBuff[0] = (byte) type.ordinal();
		//Prepare the "sourceIP" buffer
		byte[] sourceIPBuff = (sourceIPPort!=null) ? sourceIPPort.getAddress().getAddress() : new byte[0];
		//Prepare the "sourcePort" buffer 
		byte[] sourcePortBuff = (sourceIPPort!=null) ? util.Bin.intToWord(sourceIPPort.getPort()) : new byte[0];
		//return the prologue buffer
		return util.Bin.concatBuffers(msgFlag,totalSizeBuff,typeCodeBuff,sourceIPBuff,sourcePortBuff);
	}
	
}
