
package ethernet;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

/**
 * 
 * This Message extension handles messages wrapping other messages.
 * It encapsulates an "originalMessage" together with a list of destination ports
 * where the originalMessage must be redirected. This functionality is provided by the "ProxyManager" class.<p>
 *  
 * Header:<p>
 * [Number_of_ports (2 bytes)] + [ [Port_Number (2 byte)] * Number_of_ports ]<p>
 * Payload: <p>
 * the complete originalMessage (prologue+header+payload)<p>
 * 
 * @author Jose Simo (c) ai2-UPV 2014
 */
public class MessageProxyWrapper extends Message {
	
	Message originalMessage = null;
	int[] destPortsList = null;
	
	/**
	 * OriginalMessage getter.
	 * @return original message
	 */
	public Message getOriginalMessage() {
		return originalMessage;
	}

	/**
	 * Destination ports getter
	 * @return destination ports array
	 */
	public int[] getDestPortsList() {
		return destPortsList;
	}
	
	/**
	 * Constructor.
	 */
	public MessageProxyWrapper(IPPort srcIPP, Message msg,int[] destPortsLst) {
		type = Message.MessageType.PROXYWRAPPER;
		sourceIPPort = srcIPP;
		originalMessage = msg;
		destPortsList = destPortsLst;
		timeStamp = util.Clock.currentMicros();		
		bufferPayload = null; //Do not use this buffer!!! Instead, use the "originalMessage" 
		if (destPortsList != null) bufferHeader = this.computeHeader();
		if (originalMessage != null) bufferPrologue = this.computePrologue();
	}

	/**
	 * Creates an empty message to be populated using the 
	 * "fromInputStream" method.
	 */
	public MessageProxyWrapper() {
		this(null,null,null);
	}
	

	@Override
	public boolean fromInputStream(InputStream is, int size) throws IOException {
		boolean retval = true;
		int nRead = 0;
		byte[] destPortsLengthBuff = new byte[2];
		byte[] currentDestPortBuff = new byte[2];
		ByteArrayOutputStream headerStream = new ByteArrayOutputStream();
		///Header
		//Number of destinations 
		nRead = is.read(destPortsLengthBuff, 0, 2);
		if (nRead == -1) return false;
		int destPortsLength = util.Bin.wordToInt(destPortsLengthBuff);
		headerStream.write(destPortsLengthBuff);
		//Destinations array
		destPortsList = new int[destPortsLength];
		for (int i=0; i<destPortsLength; i++) {
			//Destination list element IPPort
			nRead = is.read(currentDestPortBuff, 0, 2);
			if (nRead == -1) return false;
			destPortsList[i] = util.Bin.wordToInt(currentDestPortBuff);
			headerStream.write(currentDestPortBuff);
		}
		bufferHeader = headerStream.toByteArray();
		//Payload
		bufferPayload = null; //Do not use this buffer!!! Instead, use the "originalMessage"
		originalMessage = Message.fromInputStream(is);
		
		return retval;
	}

	@Override
	protected void toOutputStream(OutputStream os) throws IOException {
		//write the message prologue (if it exists)
		if (bufferPrologue != null) {
			os.write(bufferPrologue, 0, bufferPrologue.length);
		}
		//write the message header (if it exists)
		if (bufferHeader != null) {
			os.write(bufferHeader, 0, bufferHeader.length);
		}
		//write the message payload (if it exists)
		if (originalMessage != null) {
			originalMessage.toOutputStream(os);
		}
	}
	
	/**
	 * Compute the header buffer
	 * @return the header buffer.
	 */
	private byte[] computeHeader() {
		byte[] retval = null;
		ByteArrayOutputStream headerStream = new ByteArrayOutputStream();
		//Size of destinations array	
		byte[] destLengthBuff = util.Bin.intToWord(destPortsList.length);
		//Write into the buffer the size of the array of ports destinations.
		headerStream.write(destLengthBuff, 0, 2);
		//Destinations Ip's and ports
		for (int i=0; i<destPortsList.length; i++) {		
			byte[] currentDestPortBuff = util.Bin.intToWord(destPortsList[i]);
			headerStream.write(currentDestPortBuff, 0, 2);
		}
		//Extract the buffer.
		retval = headerStream.toByteArray();
		return retval;
	}
	
	/**
	 * Override the helper Message method in order to consider the originalMessage buffers 
	 * and ignore the "MessageProxyWrapper" pay load buffer.
	 */
	@Override
	public byte[] computePrologue() {
		//Get the message header buffer
		byte[] header = getBufferHeader();
		int headerSize = (header!=null) ? header.length : 0;
		//Compute the payload buffers length
		int payLoadSize = 0;
		if (originalMessage!=null) {
			if (originalMessage.getBufferPayload() != null) payLoadSize += originalMessage.getBufferPayload().length;
			if (originalMessage.getBufferHeader() != null) payLoadSize += originalMessage.getBufferHeader().length;
			if (originalMessage.getBufferPrologue() != null) payLoadSize += originalMessage.getBufferPrologue().length;
		}
		int totalSize = headerSize + payLoadSize;
		//Prepare the "message length" buffer
		byte[] totalSizeBuff = util.Bin.intToDWord(totalSize);
		//Prepare the "message type" buffer
		Message.MessageType type = getType();
		byte[] typeCodeBuff = new byte[1];
		typeCodeBuff[0] = (byte) type.ordinal();
		//Prepare the "sourceIP" buffer
		byte[] sourceIPBuff = sourceIPPort.getAddress().getAddress();
		//Prepare the "sourcePort" buffer 
		byte[] sourcePortBuff = util.Bin.intToWord(sourceIPPort.getPort());
		//return the prologue buffer
		return util.Bin.concatBuffers(Message.msgFlag,totalSizeBuff,typeCodeBuff,sourceIPBuff,sourcePortBuff);
	}
}
