
package ethernet;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

/**
 * This Message extension handles messages for time synchronization.
 * The message contents is a long integer (8 bytes) representing 
 * a clock count (typically in microseconds).
 * This message is used in the synchronization service "SyncClock" (MultiPeerListener)
 * to synchronize the singleton clock (util.Clock)
 * Note that the "enclosed time" is different than the "time stamp".
 * The header is empty.
 * 
 * @author Jose Simo (c) ai2-UPV 2014
 * 
 */
public class MessageSyncTime extends Message {
	
	private long enclosedTime = 0;
	/**
	 * Get the time mark enclosed into the message.
	 * @return the time enclosed
	 */
	public long getEnclosedTime() {
		return enclosedTime;
	}

	/**
	 * Create a time message based on the passed buffer.
	 * @param buf message content
	 */
	public MessageSyncTime(IPPort srcIPP, byte[] buf) {
		bufferHeader = null;
		bufferPayload = buf;
		sourceIPPort = srcIPP;
		timeStamp = util.Clock.currentMicros();
		type = Message.MessageType.SYNCTIME;
		if (buf == null) {
			bufferPayload = util.Bin.longToQWord(timeStamp);
		}
		bufferPrologue = computePrologue();
	}
	/**
	 * Creates an empty message to be populated using the 
	 * "fromInputStream" method.
	 */
	public MessageSyncTime() {
		this(null,null);
	}
	/**
	 * Populates a time message reading data from an input stream.
	 */
	protected boolean fromInputStream(InputStream is, int size) throws IOException {
		boolean retval = true;
		int nRead = 0;
		byte[] data = new byte[size];
		nRead = is.read(data, 0, data.length);
		if (nRead == -1) retval = false;
		bufferPayload = data;
		bufferHeader = null;
		enclosedTime = util.Bin.qWordToLong(data);
		return retval;
	}
	
	@Override
	protected void toOutputStream(OutputStream os) throws IOException {
		//write the message prologue (if it exists)
		if (bufferPrologue != null) {
			os.write(bufferPrologue, 0, bufferPrologue.length);
		}
		//write the message header (if it exists)
		if (bufferHeader != null) {
			os.write(bufferHeader, 0, bufferHeader.length);
		}
		//write the message payload (if it exists)
		if (bufferPayload != null) {
			os.write(bufferPayload, 0, bufferPayload.length);
		}
	}
}
