
package ethernet;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

/**
 * This Message extension handles messages for data managed 
 * inside the DCPS service ("TopiscScope" MultiPeerListener). <p>
 * The message contents the TimeStamp, the Topic name and the "pay load".<p>
 * Header:<p>
 * [TimeStamp (8 bytes)] + [TopicNameSize (2 byte)] + [TopicName (TopicNameSize bytes)]<p>
 * Payload: <p>
 * plan buffer containing the data assigned to the topic.<p>
 * 
 * @author Jose Simo (c) ai2-UPV 2014
 *  
 */
public class MessageTopicData extends Message {
	
	private String topicName = null;

	public String getTopicName() {
		return topicName;
	}

	/**
	 * Create a topic message based on the passed buffer.
	 * @param nam topic name
	 * @param bufPay topic payload
	 */
	public MessageTopicData(IPPort srcIPP, String nam, byte[] bufPay) {
		type = Message.MessageType.TOPICDATA;
		sourceIPPort = srcIPP;
		topicName = nam;
		timeStamp = util.Clock.currentMicros();
		bufferPayload = bufPay;
		bufferHeader = computeHeader();
		bufferPrologue = computePrologue();
		
	}

	/**
	 * Creates an empty message to be populated using the 
	 * "fromInputStream" method.
	 */
	public MessageTopicData() {
		this(null,null,null);
	}
	
	/**
	 * Populates a topic data message reading data from an input stream.
	 * It sets the timestamp property with the message header content,
	 * propagating in this way the original message creation time.
	 */
	protected boolean fromInputStream(InputStream is, int size) throws IOException{
		boolean retval = true;
		int nRead = 0;
		byte[] nameSizeBuff = new byte[2];
		byte[] timeStampBuff = new byte[8];
		///Header
		//time stamp
		nRead = is.read(timeStampBuff, 0, 8);
		if (nRead == -1) return false;
		timeStamp = util.Bin.qWordToLong(timeStampBuff);
		//Topic name size
	    nRead = is.read(nameSizeBuff, 0, 2);
	    if (nRead == -1) return false;
		int nameSize = util.Bin.wordToInt(nameSizeBuff);
		//Topic name
		byte[] nameBuff = new byte[nameSize];
		nRead = is.read(nameBuff, 0, nameBuff.length);		
		if (nRead == -1) return false;
		topicName = new String(nameBuff);
		//Compose the header
		bufferHeader = util.Bin.concatBuffers(timeStampBuff,nameSizeBuff,nameBuff);
		//Payload
		int payloadLength = size-2-nameSize-8;	// "original size" - "header size"	
		byte[] payLoadData = new byte[payloadLength];
		nRead = is.read(payLoadData, 0, payLoadData.length);
		if (nRead == -1) return false;
		bufferPayload = payLoadData;
		bufferPrologue = computePrologue();
		return retval;
	}

	@Override
	protected void toOutputStream(OutputStream os) throws IOException {
		//write the message prologue (if it exists)
		if (bufferPrologue != null) {
			os.write(bufferPrologue, 0, bufferPrologue.length);
		}
		//write the message header (if it exists)
		if (bufferHeader != null) {
			os.write(bufferHeader, 0, bufferHeader.length);
		}
		//write the message payload (if it exists)
		if (bufferPayload != null) {
			os.write(bufferPayload, 0, bufferPayload.length);
		}
		/*
		System.out.println("OS Message type: " + getType());
		System.out.println("OS Message source: " + getSourceIPPort().toString());
		System.out.println("OS Message PRSize: " + ((getBufferPrologue() != null) ? getBufferPrologue().length : 0) );
		System.out.println("OS Message HDSize: " + ( (getBufferHeader() != null) ? getBufferHeader().length: 0) );
		System.out.println("OS Message PLSize: " + ( (getBufferPayload() != null) ? getBufferPayload().length : 0));
		*/
	}
	
	/**
	 * Compute the header of a brand new message. 
	 * It is called from the constructor. 
	 * @return the header buffer
	 */
	private byte[] computeHeader() {
		byte[] retval = null;
		byte[] timeStampBuff = util.Bin.longToQWord(timeStamp);
		if (topicName != null) {
			byte[] nameBuff = topicName.getBytes();
			int nameSize = nameBuff.length;
			retval = util.Bin.concatBuffers(timeStampBuff,util.Bin.intToWord(nameSize),nameBuff);
		} else {
			retval = timeStampBuff;
		}
		return retval;
	}

}
