
package ethernet;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;

/**
 * This Message extension handles messages for commands to 
 * configure the DCPS service ("TopiscScope" MultiPeerListener).<p>
 * The message contents the CommandType, and the "pay load" of the command.
 * For SUSCRIBE and UNSUSCRIBE commands, the payload is and array of Strings.<p>
 * Header:<p>
 *     [CommandType ordinal (1 bytes)]<p>
 * Payload: <p>
 *     [ *[ [TopicName size (2 bytes)] + [TopicName (TopicNameSize bytes)] ] ]<p>
 *     
 * @author Jose Simo (c) ai2-UPV 2014
 */
public class MessageTopicLinks extends Message {

	public enum CommandType {ASKTOPICS,SUSCRIBE,UNSUSCRIBE};
	
	private ArrayList<String> topics = null;
	private CommandType command = null;
	
	public ArrayList<String> getTopics() {return topics;}
	public CommandType getCommand() {return command;}

	/**
	 * Create a topic links message based on the passed buffer.
	 * @param cmd
	 * @param topicList
	 */
	public MessageTopicLinks(IPPort srcIPP, CommandType cmd, ArrayList<String> topicList) {
		type = Message.MessageType.TOPICLINKS;
		sourceIPPort = srcIPP;
		topics = topicList;
		command = cmd;
		timeStamp = util.Clock.currentMicros();
		if (topicList != null) bufferPayload = computePayLoad();
		if (command != null) bufferHeader = computeHeader();
		bufferPrologue = computePrologue();
	}
	
	/**
	 * Creates an empty message to be populated using the 
	 * "fromInputStream" method.
	 */
	public MessageTopicLinks() {
		this(null,null,null);
	}
	
	/**
	 * Populates a topic links message reading data from an input stream.
	 */
	@Override
	protected boolean fromInputStream(InputStream is, int size) throws IOException {

		boolean retval = true;
		int nRead = 0;
		int countRead = 0;
		byte[] commandBuff = new byte[1];
		
		//Header
		nRead = is.read(commandBuff, 0, 1);
		if (nRead == -1) return false;
		countRead += nRead;
		bufferHeader = commandBuff;
		try {
			command = CommandType.values()[(int)(commandBuff[0])];
		} catch(ArrayIndexOutOfBoundsException e) {
			throw new IOException();
		}
		//Payload
		topics = new ArrayList<String>();
		ByteArrayOutputStream os = new ByteArrayOutputStream();
		while (countRead < size) {
			//Topic name size
			byte[] nameSizeBuff = new byte[2];
			nRead = is.read(nameSizeBuff, 0, 2);
		    if (nRead == -1) return false;
			countRead += nRead;
			os.write(nameSizeBuff);
			//Topic name
			int nameSize = util.Bin.wordToInt(nameSizeBuff);
			byte[] nameBuff = new byte[nameSize];
			nRead = is.read(nameBuff, 0, nameBuff.length);		
			if (nRead == -1) return false;
			countRead += nRead;
			os.write(nameBuff);
			//Add topic name to the list
			topics.add(new String(nameBuff));
		}
		bufferPayload = os.toByteArray();
		return retval;
	}

	@Override
	protected void toOutputStream(OutputStream os) throws IOException {
		//write the message prologue (if it exists)
		if (bufferPrologue != null) {
			os.write(bufferPrologue, 0, bufferPrologue.length);
		}
		//write the message header (if it exists)
		if (bufferHeader != null) {
			os.write(bufferHeader, 0, bufferHeader.length);
		}
		//write the message payload (if it exists)
		if (bufferPayload != null) {
			os.write(bufferPayload, 0, bufferPayload.length);
		}
	}
	
	/**
	 * Compute the pay load of a brand new message. 
	 * It is called from the constructor. 
	 * @return the pay load buffer
	 */
	private byte[] computePayLoad() {
		byte[] retval = null;
		ByteArrayOutputStream os = new ByteArrayOutputStream();	
		for (int i=0; i<topics.size(); i++) {
			byte[] nameBuff = topics.get(i).getBytes();
			byte[] nameSizeBuff = util.Bin.intToWord(nameBuff.length); //two bytes
			try {
				os.write(nameSizeBuff);
				os.write(nameBuff);
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
		retval = os.toByteArray();		
		return retval;
	}
	
	/**
	 * Compute the header of a brand new message. 
	 * It is called from the constructor. 
	 * @return the header buffer
	 */
	private byte[] computeHeader() {
		byte[] retval = new byte[1];
		retval[0] = (byte) command.ordinal();
		return retval;
	}
	
	/*
	public static void main(String[] args) {
		ArrayList<String> topicList = new ArrayList<String>();
		topicList.add(new String("membo"));
		topicList.add(new String("tumbo"));
		topicList.add(new String("chory"));
		topicList.add(new String("cuely"));
		
		MessageTopicLinks msg = new MessageTopicLinks(MessageTopicLinks.LinkCommand.SUSCRIBE, topicList);
		
		ByteArrayOutputStream os = new ByteArrayOutputStream();
		try {
			os.write(msg.getBufferHeader());
			os.write(msg.getBufferPayload());
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		byte[] buff = os.toByteArray();
		ByteArrayInputStream is = new ByteArrayInputStream(buff);
		
		MessageTopicLinks msg1 = new MessageTopicLinks();
		try {
			msg1.fromInputStream(is, buff.length);
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		for (int i=0; i<msg1.getTopics().size();i++) {
			System.out.println("Topic: "+ msg1.getTopics().get(i));
		}
		
		System.out.println("command: "+ msg1.getCommand());
		
	}
	*/
	
}
