package ethernet;

import java.io.IOException;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.ArrayList;
import java.util.HashMap;

/**
 * The MultiPeer object holds a set of TCP/IP connections to support 
 * diffusion of "Messages" in a party of destinations.
 * It "accept" and "connect" to peers in an ordered way so that there is 
 * one, and only one, bidirectional link between each two peers.
 * It uses a multicast-based discovering service.
 * It propagates the received messages to listeners attending to the 
 * message type (Message.MessageType).<p>
 * <p>
 * Example of use:<p>
 * <p>
 *  public class MultiPeerTest implements MultiPeerListener {
 *	<p>
 *	public void onMessage(Message m) {...}<p>
 *  }<p>
 *  ...<p>
 *  MultiPeer mp = new MultiPeer();<p>
 *  MultiPeerTest sensor = new MultiPeerTest();<p>
 *  mp.addGeneralListener(sensor);<p>
 *  mp.send(new MessagePlain("Hola".getBytes()));<p>
 *  ...<p>
 *  <p>
 * 
 * @author Jose Simo (c) ai2-UPV 2014 <p>
 */
public class MultiPeer {
	
	//Define the operation options
	private static boolean OPTION_SYNCCLOCKS = true;
	private static boolean OPTION_SENDBYPROXY = true;
	///////	
	//Multicast "type D" address: from 224.0.0.0 to 239.255.255.255
	private static String defaultAddressMulticast = "228.10.1.1"; 
	//Any port to identify the Multicast channel.
	private static int defaultPortMulticast = 7110;
	//Port to accept TCP/IP Unicast connections.
	private static int defaultPortServer = 5000;
	
	private ArrayList<MultiPeerListener> generalListeners = null;
	private HashMap<Message.MessageType,ArrayList<MultiPeerListener>> specificListeners = null;
	//Destinations organized by IP (1st) and port (2nd)
	private PeersSet destinationsSet = null;
	//Local identity and server port
	private IPPort localServiceIPPort = null;
	private ServerSocket servSock = null;
	//Unique incoming messages queue
	private QueueMessage receivedMessages = new QueueMessage();
	
	public IPPort getLocalIPPort() { return localServiceIPPort;}
	
	/**
	 * Create a customized "MultiPeer" object calling the initialization method..
	 * @param initialPort (int) is the intended local port to accept connections.
	 * @param ipMulticast (String) is the IP for the multicast discovering service.
	 * @param portMulticast (int) is the port for the multicast discovering service.
	 */
	public MultiPeer(int initialPort, String ipMulticast, int portMulticast) {
		InitMultiPeer( initialPort,  ipMulticast, portMulticast);
	}
	
	/**
	 * Create a customized "MultiPeer" object with the default multicast discovering channel.
	 * @param initialPort (int) is the intended port to accept connections.
	 */
	public MultiPeer(int initialPort) {
		this(initialPort,defaultAddressMulticast,defaultPortMulticast);
	}
	
	/**
	 * Create a default "MultiPeer" object (default "initialPort" and multicast discovering channel).
	 */
	public MultiPeer() {
		this(defaultPortServer,defaultAddressMulticast,defaultPortMulticast);
	}
	
	/**
	 * MultiPeer initialization procedure.
	 * @param initialPort (int) is the intended local port to accept connections.
	 * @param ipMulticast (String) is the IP for the multicast discovering service.
	 * @param portMulticast (int) is the port for the multicast discovering service.
	 */
	public void InitMultiPeer(int initialPort, String ipMulticast, int portMulticast) {
		//Create the listeners arrays
		generalListeners = new ArrayList<MultiPeerListener>();
		specificListeners = new HashMap<Message.MessageType,ArrayList<MultiPeerListener>>();
		for (Message.MessageType mType: Message.MessageType.values()) {
			specificListeners.put(mType, new ArrayList<MultiPeerListener>());
		}
		//Set the local identity ("IPPort"). See the "utils.Net" singleton
		//to get information about how manage the case of multiple network interfaces up.
		localServiceIPPort = new IPPort(util.Net.getFavouriteLocalIP(),initialPort);
		//Set the server socket to accept connections.
		//this call can side-modify the "initialPort" to find one available. 
		boolean readyToAcceptConnections = this.setServerSocket();
		//Create the Destinations (peers in the party) collection (PeersSet)
		//The Destinations are inserted in the collection using "addDestination" 
		//methods (removed with "removeDestination") on fly. See "Destination", 
		//"ServerThread" and "DiscoveringService" code
		destinationsSet = new PeersSet(localServiceIPPort);		
		//Start connection services.
		if (readyToAcceptConnections) {
			//Start listeners callback propagation thread 
			Thread propagation = new PropagationThread();
			propagation.setPriority(Thread.NORM_PRIORITY+1);
			propagation.start();
			//Start connections server
			Thread server = new ServerThread();
			server.setPriority(Thread.MAX_PRIORITY);
			server.start();
			//Start discovering service
			new DiscoveringService(this, new IPPort(ipMulticast,portMulticast));	
		} else  {
			util.Log.logMessage(util.Log.LogType.ERROR,"MultiPeer Connection server: cannot open server socket. ");
		}
		//Start attached clock synchronization service (multiPeer listener)
	    if (OPTION_SYNCCLOCKS) new SyncClock(this);
		if (OPTION_SENDBYPROXY) new ProxyManager(this);
	}
	
	/**
	 * From the peers in the party, get "local" and "global" (network) leader attending to the IPPort ordering criteria.
	 * @return the Destinations found. Null if the corresponding leader is the local one.
	 */

	public Destination getNetworkLeader() {
		return destinationsSet.getGeneralLeader();
	}
	
	public Destination getLocalLeader() {
		return destinationsSet.getNodeLeader(this.getLocalIPPort().getIpAddressAsLong());
	}
	
	public Destination getDestinationByIPPort(IPPort ipp) {
		Destination retval = destinationsSet.getDestinationByIPPort(ipp);
		return retval;
	}
	
	
	/**
	 * Create the server socket used in "ServerThread" to accept connections.
	 * It looks for an available port starting from the value given in the constructor (max 100 tries).
	 * @return true if success.
	 */
	private boolean setServerSocket() {
		boolean retval = false;
		int numberOfAttemps = 0;
		while (numberOfAttemps < 100) {
			try {
				servSock = new ServerSocket(localServiceIPPort.getPort());
				retval = true;
				util.Log.logMessage(util.Log.LogType.INFO,"MultiPeer Connection server: Connected in port: "+ localServiceIPPort.getPort());
				break;
			} catch (IOException e) {
				util.Log.logMessage(util.Log.LogType.WARNING,"MultiPeer Connection server: port not available: "+ localServiceIPPort.getPort());
				localServiceIPPort = new IPPort(localServiceIPPort.getAddress(),localServiceIPPort.getPort()+1);
				numberOfAttemps++;
			}
		}
		return retval;
	}
	
	/**
	 * Send a Message to a subset of peers in the party.
	 * @param m
	 * @return An array containing the failed destinations.
	 */
	public Destination[] send(Destination[] destArray, Message m) {
		//Create an auxiliary PeersSet and populate it with the target destinations.
		PeersSet auxDestSet = new PeersSet(this.getLocalIPPort());
		for (Destination dest: destArray) {
			auxDestSet.addDestination(dest);
		}
		//Send
		Destination[] retval = null;
		if (OPTION_SENDBYPROXY) {
			retval = sendByProxyPeersSet(m,auxDestSet);
		} else {
			retval = sendDirectPeersSet(m,auxDestSet);
		}
		this.propagateNewMessage(m);
		return retval;
	}
	
	/**
	 * Send a Message to ALL peers in the party.
	 * @param m
	 */
	public void send(Message m) { 
		//Send
		if (OPTION_SENDBYPROXY) {
			sendByProxyPeersSet(m,destinationsSet);
		} else {
			sendDirectPeersSet(m,destinationsSet);
		}
		this.propagateNewMessage(m);
	}
	
	/**
	 * Auxiliary method to send a message to an arbitrary "PeersSet"
	 * using the "ProxyManager" service. It only send network messages
	 * to the "leader" of each network node. The local "ProxyManager" 
	 * propagates the message to the node local peers.
	 * The failed peers (Destination) are removed from "destinationsSet" collection.
	 * @param m
	 * @param pSet
	 * @return an array containing the failed destinations.
	 */
	private Destination[] sendByProxyPeersSet(Message m, PeersSet pSet) { 
		ArrayList<Destination> toRemove = new ArrayList<Destination>();
		//Send the message to ALL local destinations (same node)
		Destination[] toSendMessage = pSet.getLocalDestinations();
		if (toSendMessage != null) {
			for(Destination dest:toSendMessage) {
				boolean success = dest.send(m);
				if (!success) {
					toRemove.add(dest);
				}	
			}
		}	
		//Send a "MessageProxyWrapper" message to the peer leader of each (remote) network node.
		long[] listIPs = pSet.getPeersSetIPs();
		for (long ip: listIPs) {
			if (ip != pSet.getLocalNode().getIpAddressAsLong()) { //Consider only remote nodes.
				Destination leader = pSet.getNodeLeader(ip);
				if (leader != null) {
					MessageProxyWrapper proxyMsg = new MessageProxyWrapper(this.getLocalIPPort(),m,pSet.getRemoteNoLeaderDestPorts(ip));
					boolean success = leader.send(proxyMsg);
					if (!success) {
						toRemove.add(leader);
					}	
				}
			}	
		}	
		//Remove the failed Destinations.
		for (Destination dest:toRemove) {
			removeDestination(dest);
		}
		return toRemove.toArray(new Destination[0]);
	}
	
	/**
	 * Auxiliary method to send a message to an arbitrary "PeersSet" directly.
	 * It generates network messages for ALL peers in the set. 
	 * It does not use the "ProxyManager" service.
	 * The failed peers (Destination) are removed from "destinationsSet" collection.
	 * @param m
	 * @param pSet
	 * @return an array containing the failed destinations.
	 */
	private Destination[] sendDirectPeersSet(Message m, PeersSet pSet) { 	
		ArrayList<Destination> toRemove = new ArrayList<Destination>();
		//Send the message to ALL destinations in the set.
		long[] listIPs = pSet.getPeersSetIPs();
		for (long ip: listIPs) {
			Destination[] destList = pSet.getIPDestinations(ip);
			if (destList != null) {
				for(Destination dest:destList) {
					boolean success = dest.send(m);
					if (!success) {
						toRemove.add(dest);
					}	
				}
			}
		}
		//Remove the failed Destinations.
		for (Destination dest:toRemove) {
			removeDestination(dest);
		}
		return toRemove.toArray(new Destination[0]);
	}
	
	/**
	 * Callback generated by "Destinations" when a message arrives.
	 * This method enqueues the message in a synchronized queue.
	 * @param m
	 */
	public void OnMessage(Message m) {
		//System.out.println("Received: "+new String(m.getBufferPayload()) + 
		//                   " From: "+ m.getSource().getIPPort().toString() + " Type:" + m.getType());
		receivedMessages.push(m);
	}
	
	/**
	 * Add a "MultiPeerListener" that will be called-back for all message types.
	 * @param obj
	 */
	public void addGeneralListener(MultiPeerListener obj) {
		synchronized(generalListeners) {
			generalListeners.add(obj);
		}
	}
	public void removeGeneralListener(MultiPeerListener obj) {
		synchronized(generalListeners) {
			generalListeners.remove(obj);
		}
	}
	
	/**
	 * Add a "MultiPeerListener" that will be called-back only when arrives 
	 * a message of a specific type.
	 * @param type
	 * @param obj
	 */
	public void addSpecificListener(Message.MessageType type, MultiPeerListener obj) {
		synchronized(specificListeners) {
			specificListeners.get(type).add(obj);
		}
	}
	public void removeSpecificListener(Message.MessageType type,MultiPeerListener obj) {
		synchronized(specificListeners) {
			specificListeners.get(type).remove(obj);
		}
	}
	
	/**
	 * 
	 * @param ipp
	 * @return true if contained
	 */
	public boolean isDestinationContained(IPPort ipp) {
		return destinationsSet.isDestinationContained(ipp);
	}
	
	/**
	 * Adds a destination suitable to "connect" to it. 
	 * This method only connects to Destinations "greater than" the "local" one.
	 * @param ipp The destination IPPort.
	 * @return true if the "connect" has been successful and the destination added to the party. 
	 * False if the destination is already in the party or is "lower than" the local.
	 */
	public boolean addDestination(IPPort ipp) {
		boolean retval = false;
		Destination dest = null;
	    //Decides to create the destination and connect it.
		boolean shouldConnect = ipp.isGreater(localServiceIPPort);
		boolean alreadyContained = this.isDestinationContained(ipp);
		if ( (!alreadyContained) && (shouldConnect) ) {
			dest = new Destination(this);
			boolean connected = dest.connect(ipp);
			if (connected) {
				//Add the destination to the set.
				destinationsSet.addDestination(dest);
				//Propagate the new connection to listeners (callback)
				propagateNewConection(dest);
				retval = true;
			}
		}
		return retval;
	}
	
	/**
	 * Add a destination "accepted" by the "ServerThread".
	 * @param sock
	 * @return true if successfully connected.
	 */
	public boolean addDestination(Socket sock) {
		boolean retval = false;
		Destination dest = new Destination(this);
		boolean connected = dest.accepted(sock);
		if (connected) {
			//Add the destination to the set.
			destinationsSet.addDestination(dest);
			//Propagate the new connection to listeners (callback) 
			propagateNewConection(dest);
			retval = true;
		}
		return retval;
	}
	
	/**
	 * Remove a destination from the party.
	 * Called when detects a communication fault.
	 * @param dest
	 */
	public synchronized void removeDestination(Destination dest) {
		util.Log.logMessage(util.Log.LogType.INFO,"MultiPeer: Removing destination: " + dest.getIPPort().toString());
		destinationsSet.removeDestination(dest);
	}
	
	/**
	 * Propagate the new connection to listeners (callback)
	 * @param dest
	 */
	private void propagateNewConection(Destination dest) {
		synchronized(generalListeners) {
			for (MultiPeerListener listener: generalListeners) {
				try {
					listener.onNewConnection(dest); //Modify it to do async call.
				} catch (Exception e) {
					util.Log.logMessage(util.Log.LogType.ERROR,"MultiPeer: Erron in listener onNewConnection callback: " + listener.getClass().getName());
				}
			}
		}
		synchronized(specificListeners) {
			for (Message.MessageType type : Message.MessageType.values()) {
				for (MultiPeerListener listener: specificListeners.get(type)) {
					try {
						listener.onNewConnection(dest); //Modify it to do async call.
					} catch (Exception e) {
						util.Log.logMessage(util.Log.LogType.ERROR,"MultiPeer: Erron in listener onNewConnection callback: " + listener.getClass().getName());
					}
				}
			}
		}
	}
	
	/**
	 * Propagate the new message to listeners (callback)
	 * @param m
	 */
	private void propagateNewMessage(Message m) {
		synchronized(generalListeners) {
			for (MultiPeerListener listener: generalListeners) {
				try {
					listener.onMessage(m); //Modify it to do async call.
				} catch (Exception e) {
					util.Log.logMessage(util.Log.LogType.ERROR,"MultiPeer: Erron in listener OnMessage callback: " + listener.getClass().getName());
				}
			}
		}
		synchronized(specificListeners) {
			for (MultiPeerListener listener: specificListeners.get(m.getType())) {
				try {
					listener.onMessage(m); //Modify it to do async call.
				} catch (Exception e) {
					util.Log.logMessage(util.Log.LogType.ERROR,"MultiPeer: Erron in listener OnMessage callback: " + listener.getClass().getName());
				}
			}
		}
	}
	
	/**
	 * This private thread dequeues received messages and dispatch it to the local listeners.
	 * @author jsimo
	 *
	 */
	private class PropagationThread extends Thread {
		@Override
		public void run() {			
			while(Thread.currentThread().isAlive()) {
				Message m = receivedMessages.pull();
				if (m != null) {
					//Propagate the new message to listeners (callback)
					propagateNewMessage(m);
					//System.out.println("Propagating message: " + m.getType());
					
				} 
			}
		}
	}
		
	/**
	 * This private thread runs forever accepting connections 
	 * and creating accepted Destinations in the party.
	 * @author jsimo
	 *
	 */
	private class ServerThread extends Thread {
		@Override
		public void run() {		
			while(Thread.currentThread().isAlive()) {	
				try {
					Socket s = servSock.accept();
					addDestination(s);	
				} catch (IOException e) {
					util.Log.logMessage(util.Log.LogType.ERROR,"MultiPeer: Error accepting connection. Server thread exit.");
					return;
				}
			}
		}
	}
	
}
