
package ethernet;

import java.util.HashMap;

/**
 * A "PeersNode" is a set of end points (peers, "Destination" objects).
 * A "PeersNode" is a set of peers within the same IP address. 
 * The peers are indexed by "port".<p>
 * Each "PeersNode" object has a leader node, corresponding the lowest destination according the ordering criteria.<p>
 * The leader nodes play special roles in "SyncClock" and "ProxyManager" services.
 * 
 * @author Jose Simo (c) ai2-UPV 2014
 */
public class PeersNode {
	
	long IPAsLong;
	Destination nodeLeader;
	HashMap<Integer,Destination> nodeEndPoints;
	
	/**
	 * Constructor. 
	 * @param ip
	 */
	PeersNode(long ip) {
		IPAsLong = ip;
		nodeLeader = null;
		nodeEndPoints = new HashMap<Integer,Destination>();
	}
	
	/**
	 * Checks if a port is the data structure.
	 * @param port
	 * @return true if contained.
	 */
	boolean containsPort(int port) {
		return nodeEndPoints.containsKey(port);
	}
	
	/**
	 * Adds a destination to the data structure.
	 * Refresh the "leader" of the node taking into account the new Destination.
	 * @param dest
	 * @return true if added, false if already exist.
	 */
	boolean addDestination(Destination dest) {
		boolean retval = false;
		long ip = dest.getIPPort().getIpAddressAsLong();
		int port = dest.getIPPort().getPort();
		//Check the IP
		if (IPAsLong != ip ) {
			util.Log.logMessage(util.Log.LogType.ERROR,"EndPointsSet addDestination: Wrong IP. No added. ");
			return false;
		}
		if (!nodeEndPoints.containsKey(port)) {
			nodeEndPoints.put(port, dest);
			refreshLeader();
			retval = true;
		}
		return retval; 
	}
	
	/**
	 * Removes a destination from the data structure.
	 * Refresh the "leader" of the node taking into account the removed Destination.
	 * @param dest
	 * @return true if successfully removed.
	 */
	boolean removeDestination(Destination dest) {
		boolean retval = false;
		long ip = dest.getIPPort().getIpAddressAsLong();
		int port = dest.getIPPort().getPort();
		//Check the IP
		if (IPAsLong != ip ) {
			util.Log.logMessage(util.Log.LogType.ERROR,"EndPointsSet removeDestination: Wrong IP. No removed. ");
			return false;
		}
		Object removed = nodeEndPoints.remove(port);
		if (removed != null) {
			refreshLeader();
			retval = true;
		}
		return retval;
	}
	
	/**
	 * Assigns the leader node to the lower one according the ordering criteria.
	 */
	private void refreshLeader() {
		int currentPort = Integer.MAX_VALUE;
		Destination currentLeader = null;
		for(Destination dest:nodeEndPoints.values()) {
			if (dest.getIPPort().getPort() < currentPort) {
				currentPort = dest.getIPPort().getPort();
				currentLeader = dest;
			}
		}
		nodeLeader = currentLeader;
	}
}
