
package ethernet;

import java.util.HashMap;

/**
 * It stores an organized set of "Destinations". In particular, the MultiPeer Destinations.
 * Data structure to store "Peers Nodes" set information is organized by IP address (indexed by "IPAsLong" key).<p>
 * 
 * A "PeersSet" is a set of "PeersNode" objects. A "PeersNode" is a set of end points (peers, "Destination" objects).
 * A "PeersNode" is a set of peers within the same IP address. 
 * Each "PeersNode" object has a leader node, corresponding the lowest destination according the ordering criteria.
 * The "PeersSet" general leader is the lowest among the PeersNode leaders.<p>
 * 
 * The leader nodes play special roles in "SyncClock" and "ProxyManager" services.<p>
 *
 * @author Jose Simo (c) ai2-UPV 2014
 */
public class PeersSet {
	//The local node identity is stored apart from the rest of nodes. 
	//So, the "nodes" data structure stores the identity of all nodes but the local node.
	//Observe how "MultiPeer" (by the way of "DiscoveringService") stores in the "PeerSet" 
	//data structure, only peer destinations, not the local one.
	private IPPort localNode = null;
	public IPPort getLocalNode() { return localNode;	}
	//Use Long (64bit) instead Integer (32bit) for IP representation because Integer can not be unsigned (ordering properties)
	private HashMap<Long,PeersNode> nodes;
	
	/**
	 * Constructor
	 * @param ipp the local node identity as IPPort.
	 */
	PeersSet(IPPort ipp) {
		localNode = ipp;
		nodes = new HashMap<Long,PeersNode>();
	}

	/**
	 * Adds a destination into the data structure.
	 * It controls the creation of new "PeerNode" objects.
	 * @param dest the "Destination" to store.
	 * @return true if "dest" added. False if already exist.
	 */
	public synchronized boolean addDestination(Destination dest) {
		boolean retval = false;	
		long lip = dest.getIPPort().getIpAddressAsLong();
		PeersNode currentNode = null;
		if (!nodes.containsKey(lip)) {
			currentNode = new PeersNode(lip);
			nodes.put(lip, currentNode);
		} else {
			currentNode = nodes.get(lip);
		}
		retval = currentNode.addDestination(dest);
		return retval;
	}
	
	/**
	 * Remove a destination from the data structure.
	 * It controls the destruction of empty "PeerNode" objects.
	 * @param dest the "Destination" to remove.
	 * @return true if "dest" removed successfully.
	 */
	public synchronized boolean removeDestination(Destination dest) {
		boolean retval = false;	
		long lip = dest.getIPPort().getIpAddressAsLong();
		PeersNode currentNode = nodes.get(lip);	
		if (currentNode != null) {
			retval = currentNode.removeDestination(dest);		 
			if (currentNode.nodeEndPoints != null) {
				if (currentNode.nodeEndPoints.size() == 0) {
					nodes.remove(lip);
				}
			}
		}
		return retval;
	}
	
	/**
	 * Gets the "Destination" node leader in node specified by IP (as long).
	 * @param nodeIP the IP address (long representation) of the node.
	 * @return the leader Destination reference, null if the leader is the local one.
	 */
	public synchronized Destination getNodeLeader(long nodeIP) {
		Destination retval = null;
		PeersNode currentNode = nodes.get(nodeIP);
		if (currentNode != null) {
			retval = currentNode.nodeLeader;
		} 
		//Return null if the real leader is the local node.
		if (retval != null) {
			if (nodeIP == localNode.getIpAddressAsLong()) {
				if (localNode.getPort() < retval.getIPPort().getPort()) {
					retval = null;
				}
			}
		}
		return retval;
	}
	
	/**
	 * Gets the leader among leaders.
	 * @return the leader Destination reference, null if the leader is the local one.
	 */
	public synchronized Destination getGeneralLeader() {
		Destination retval = null;
		long currentLip = Long.MAX_VALUE;
		for (PeersNode node: nodes.values()) {
			if (node.IPAsLong < currentLip) {
				currentLip = node.IPAsLong;
				retval = node.nodeLeader;
			}
		}
		//Return null if the real leader is the local node.
		if (retval != null) {
			if ( retval.getIPPort().isGreater(localNode) ) {
				retval = null;
			}
		}
		return retval;
	}
	
	/**
	 * Checks if an end point is in the data structure.
	 * @param ipp
	 * @return true if contained.
	 */
	public boolean isDestinationContained(IPPort ipp) {
		boolean retval = false;	
		Destination d = getDestinationByIPPort(ipp);
		if (d != null) retval = true;	
		return retval;
	}
	
	/**
	 * Gets a "Destination" reference from the IPPort identity.
	 * @param ipp
	 * @return the "Destination" reference if found. Null if not found.
	 */
	public synchronized Destination getDestinationByIPPort(IPPort ipp) {
		Destination retval = null;
		PeersNode currentNode = nodes.get(ipp.getIpAddressAsLong());
		if (currentNode != null) {
			retval = currentNode.nodeEndPoints.get(ipp.getPort());	
		} 
		return retval;	
	}
	
	/**
	 * Gets an array containing the "Destinations" running in the local IP address. 
	 * The local one is not included in the array.
	 * @return a Destinations array.
	 */
	public synchronized Destination[] getLocalDestinations() {
		return getIPDestinations(localNode.getIpAddressAsLong());
	}
	
	/**
	 * Gets an array containing the "Destinations" running in a specific IP address. 
	 * @return a Destinations array.
	 */
	public synchronized Destination[] getIPDestinations(long ip) {
		Destination[] retval = null;
		PeersNode netPeers = nodes.get(ip);
		if (netPeers != null) {
			int dim = netPeers.nodeEndPoints.size();
			retval = new Destination[dim];
			int i = 0;
			for (Destination dest:netPeers.nodeEndPoints.values()) {
				retval[i] = dest;
				i++;
			}
		}
		return retval;
	}
	
	/**
	 * Gets an array of the IP's in the party.
	 * @return an IP-as-long array.
	 */
	public synchronized long[] getPeersSetIPs() {
		long[] retval = null;
		retval = new long[nodes.keySet().size()];
		int i = 0;
		for (long ip:nodes.keySet()) {
			retval[i] = ip;
			i++;
		}
		return retval;
	}
	
	/**
	 * Gets an array of remote destinations (in the specified IP) not including the leader one.
	 * It is an auxiliary method used to compose a valid "ProxyWrapperMessage".
	 * @param ipAsLong
	 * @return an array of integers (dest ports)
	 */
	public synchronized int[] getRemoteNoLeaderDestPorts(long ipAsLong) {
		int[] retval = null;
		//ensure ipAsLong is not the local ip
		if (ipAsLong != localNode.getIpAddressAsLong()) { 
			//Get the node peers.
			PeersNode remotePeers = nodes.get(ipAsLong);
			if (remotePeers != null) {
				int dim = remotePeers.nodeEndPoints.size();
				retval = new int[dim - 1];
				int i = 0;
				for (Destination dest:remotePeers.nodeEndPoints.values()) {
					if (!dest.getIPPort().isEqual(remotePeers.nodeLeader.getIPPort())) {
						//Store the peer port but the leader port.
						retval[i] = dest.getIPPort().getPort();
						i++;
					}
				}
			}
		}
		return retval;
	}
	
}
