
package ethernet;

/**
 * This class implements the Cristian's clock synchronization protocol.
 * It is a "MultiPeerListener" intended to handle the messages of "SYNCTIME" type.
 * The target clock (clock to synchronize) is the "util.Clock" singleton.<p>
 * 
 * This class acts as time-server or time-client. The "SyncronizationThread" send
 * "0L" messages periodically to the chosen server in order to get the server time stamp
 * and measure the round trip time. The "0L" message is answered directly inside 
 * the "OnMessage" method (acting in the server side).<p>
 * Because the related "MultiPeer" is provided in construction time, the newly created 
 * object registers itself as MultiPeer "Specific Listener".<p>
 * 
 * The period of synchronization is shorter during startup or when the correcting 
 * offset exceeds the limit value, getting slower during normal operation.<p>
 * 
 * @author Jose Simo (c) ai2-UPV 2014
 */
public class SyncClock implements MultiPeerListener {
	//Maximum synchronization period
	private static int messagePeriodMAX = 10000;  //milliseconds
	//Minimum synchronization period
	private static int messagePeriodMIN = 100;    //milliseconds
	//Correcting offset limit
	private static int maxAllowedOffset = 100;    //microseconds
	//Current synchronization period
	private int messagePeriod = messagePeriodMIN;

	private MultiPeer netPeers = null;
	
	private long tSend = 0;
	private long tRecv = 0;
	private long tServer = 0;
	private boolean messageReady = false;
	
	/**
	 * Constructor. The associated MultiPeer must be provided.
	 * The newly created object registers itself as MultiPeer "Specific Listener".
	 * It creates and starts a new "SynchronizationThread".
	 * @param mp
	 */
	public SyncClock(MultiPeer mp) {
		netPeers = mp;
		netPeers.addSpecificListener(Message.MessageType.SYNCTIME, this);
		new SynchronizationThread().start();
	}
	/**
	 * MultiPeerListener callback.
	 */
	@Override
	public void onNewConnection(Destination dest) {
		messagePeriod = messagePeriodMIN;
	}
	/**
	 * MultiPeerListener callback.
	 */
	@Override
	public synchronized void onMessage(Message m) {
		//Gets the message contents
		long receivedtServer = util.Bin.qWordToLong(m.getBufferPayload());
		//If the message is 0L, acts as a time server
		if (receivedtServer == 0L) {
			//Answer the current local clock
			Destination ckDest = netPeers.getDestinationByIPPort(m.getSourceIPPort());
			MessageSyncTime ansMsg = new MessageSyncTime(netPeers.getLocalIPPort(), util.Bin.longToQWord(util.Clock.currentMicros()));
			ckDest.send(ansMsg);
		} else { //Acts as a client
			//Store the server response relevant information
			tRecv = util.Clock.currentMicros();
			tServer = receivedtServer;
			//Now the server response is ready. Change the state and notify.
			messageReady = true;
			notify();
		}	
	}
	
	/**
	 * Used in "SynchronizationThread" to wait until the time server response.
	 * Wait until response message ready.
	 * It gives to the server 100ms to complete the response (timeout).
	 * If this time expires, stops the wait and resets the cycle (message not ready)
	 */
	private synchronized void waitForServerResponse() {
		while (!messageReady) {
			try {
				util.Clock.Chronometer chrono = new util.Clock.Chronometer();
				wait(100);
				if (chrono.diffNow() >= 100) { break; }
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
		messageReady = false;
	}
	/**
	 * @author jsimo
	 * It is the "client side" synchronization activity.
	 * Do periodically 10 time server interactions and use the best one to synchronize the clock.
	 * The period of synchronization is shorter during startup or when the correcting 
	 * offset exceeds the limit value, getting slower during normal operation.
	 */
	private class SynchronizationThread extends Thread {
		@Override
		public void run() {			
			int nSamples = 10;
			SyncSample[] samples = new SyncSample[nSamples];
			messagePeriod = messagePeriodMIN;
			while(Thread.currentThread().isAlive()) {
				//Choose the time server peer. If the decision is null, it is the node master clock.
				Destination timeServerDest = netPeers.getLocalLeader();
				//If I'm the local leader, look for the network leader.
				if (timeServerDest == null){
					timeServerDest = netPeers.getNetworkLeader();
				}
				//If decided the sync server, do it:
				if (timeServerDest != null) {
					//System.out.println("Sync to: " + timeServerDest.getIPPort().toString());
					//Do "nSamples" sync messages and store the results
					synchronized(SyncClock.this) {
						for (int i=0; i<nSamples; i++) {
							tSend = util.Clock.currentMicros();
							//Send 0L to start the synchronization
							timeServerDest.send(new MessageSyncTime(netPeers.getLocalIPPort(), util.Bin.longToQWord(0L))); 
						    //wait response
							waitForServerResponse(); 
							//Compute sync. params. and store it into the array
							long roundTrip = tRecv-tSend;
							long offset =  tServer - tRecv +roundTrip/2L;
							samples[i] = new SyncSample(offset,roundTrip);
						}
					}	
					//Uses the best try to sync the clock.
					long minRT = Long.MAX_VALUE;
					int minRTIndx = 0;
					for (int i=0; i< nSamples ; i++) {
						//Chose the best discarding the abnormal negative round trip times.
						if ( (samples[i].roundTrip > 0) && (samples[i].roundTrip < minRT) ) {
							minRT = samples[i].roundTrip;
							minRTIndx = i;
						}
					}
					//If the best candidate is not abnormal, apply the correction
					if (samples[minRTIndx].roundTrip >= 0) {
						util.Clock.setOffset(samples[minRTIndx].offset, samples[minRTIndx].roundTrip / 2L);
					}
					//Decide the sleep time:
					//If out-of-sync, set the minimum sleep time. If well sync, increase the period
					if ( samples[minRTIndx].offset > maxAllowedOffset) {
						messagePeriod = messagePeriodMIN;
					} else {
						messagePeriod = (messagePeriod >= messagePeriodMAX) ? messagePeriodMAX: messagePeriod + messagePeriod/2;
					}
				} else {
					//If the object is the master clock, check this role every 500ms.
					messagePeriod = 500;
					//System.out.println("Sync: This is time server");
				}
				//Sleep until next sync,
				try {
					Thread.sleep(messagePeriod);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
		}
	}

	/**
	 * Auxiliary class to store the sync tries results.
	 * @author jsimo
	 *
	 */
	public static class SyncSample {
		public long offset = 0;
		public long roundTrip = 0;
		SyncSample(long o, long r) {
			offset = o;
			roundTrip = r;
		}
	}

	
}
