
package util;

/**
 * Clock utilities:
 * It implements an static clock (singleton) supporting synchronization offset.
 * This static clock is intended to time stamp events (in microseconds).
 * Some external synchronization procedure must call the "setOffset" method
 * in order to have synchronized time stamps.
 * 
 * Also included:
 * Class: Chronometer -> to measure time differences
 * Class: WatchDog -> Very simple (and inefficient) watch dog implementation.
 * (No use it to fine grained or low level purposes)
 * 
 * @author Jose Simo (c) ai2-UPV 2014
 */

public class Clock {
	private static long offset = 0;
	private static long syncError = 0;
	/**
	 * Get the clock value:
	 * @return the time stamp value in microseconds.
	 */
	public static long currentMicros() {
		return System.nanoTime() / 1000 + offset;
	}
	/**
	 * Set the differential offset (and estimated error) 
	 * to correct the clock synchronization.
	 * @param of The offset in microseconds
	 * @param err The absolute error in microseconds
	 */
	public static void setOffset(long of, long err) {
		offset += of;
		syncError = err;
	}
	/**
	 * Clock offset getter.
	 * @return the current offset in microseconds
	 */
	public static long getOffset() { return offset;}
	/**
	 * Clock error getter
	 * @return the current error
	 */
	public static long getError() { return syncError;}
	
	/**
	 * @author jsimo
	 * Chronometer to measure time differences
	 */
	public static class Chronometer {	
		private long startTime = 0;
		public Chronometer() {start();}
		public void start() { startTime = System.nanoTime(); }
		public void stop() { startTime = 0L; }
		public long diffNow() { return (System.nanoTime() - startTime) / 1000;}
	}
	
	/**
	 * @author jsimo
	 * Very simple (and inefficient) watch dog implementation.
	 * (No use it to fine grained or low level purposes)
	 */
	public static class WatchDog {
		private static long resolution = 20;
		private long fireCountDown;
		public WatchDog(long milis) { fireCountDown = milis; }
		public void kick(long milis) { fireCountDown = milis; }
		public void waitBark() {
			while (fireCountDown > 0) {
				try {
					Thread.sleep(resolution);
				} catch (InterruptedException e) {/*nop*/}
				fireCountDown -= resolution;
			}
		}
	}

}
